import numpy as np
from PIL import Image
import os
import cupy
import chainer.cuda


chanel_num = 3
layer_num = 3
view_point_row_num = 5
view_point_num = view_point_row_num*view_point_row_num
pad_size = (int)(view_point_row_num/2)
iteration = 50

if os.path.isdir("./Dst") == False:
	os.mkdir("./Dst")

if os.path.isdir("./Dst/layer") == False:
	os.mkdir("./Dst/layer")

if os.path.isdir("./Dst/reproducedimage") == False:
	os.mkdir("./Dst/reproducedimage")



#Load light field datas
cpu_light_field = []
cpu_layer = []
cpu_reproduced_light_field = []

for t in range(view_point_num):
	fname = "../../LightField/dice%02d.png" %(t+1)
	img_load = Image.open(fname)
	img_load = np.asarray(img_load)
	img_load = img_load.astype(np.float32) / 255.0
	cpu_light_field.append(img_load.copy())

size_h = cpu_light_field[0].shape[0]
size_w = cpu_light_field[0].shape[1]

for t in range(layer_num):
	cpu_layer.append(np.zeros((size_h, size_w, chanel_num)))

for t in range(view_point_num):
	cpu_reproduced_light_field.append(np.zeros((size_h, size_w, chanel_num)))


#Send light field to GPU
gpu_light_field = np.zeros((chanel_num, view_point_num, size_h+pad_size*4, size_w+pad_size*4), np.float32)

for ch in range(chanel_num):
	for t in range(view_point_num):
		gpu_light_field[ch, t, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = cpu_light_field[t][:, :, ch]

gpu_light_field = chainer.cuda.to_gpu(gpu_light_field)


#Layers initialization
numerator = cupy.zeros((chanel_num, size_h, size_w), cupy.float32)
denominator = cupy.zeros((chanel_num, size_h, size_w), cupy.float32)
gpu_reproduced_light_field = cupy.zeros((chanel_num, view_point_num, size_h+pad_size*4, size_w+pad_size*4), cupy.float32)

gpu_front_layer = cupy.random.rand(chanel_num, size_h+pad_size*4, size_w+pad_size*4)
gpu_middle_layer = cupy.random.rand(chanel_num, size_h+pad_size*4, size_w+pad_size*4)
gpu_rear_layer = cupy.random.rand(chanel_num, size_h+pad_size*4, size_w+pad_size*4)

gpu_front_layer[:, 0:pad_size*2, 0:size_w+pad_size*4] = 0.0
gpu_front_layer[:, 0:size_h+pad_size*4, 0:pad_size*2] = 0.0
gpu_front_layer[:, 0:size_h+pad_size*4, size_w+pad_size*2:size_w+pad_size*4] = 0.0
gpu_front_layer[:, size_h+pad_size*2:size_h+pad_size*4, 0:size_w+pad_size*4] = 0.0

gpu_middle_layer[:, 0:pad_size*2, 0:size_w+pad_size*4] = 0.0
gpu_middle_layer[:, 0:size_h+pad_size*4, 0:pad_size*2] = 0.0
gpu_middle_layer[:, 0:size_h+pad_size*4, size_w+pad_size*2:size_w+pad_size*4] = 0.0
gpu_middle_layer[:, size_h+pad_size*2:size_h+pad_size*4, 0:size_w+pad_size*4] = 0.0

gpu_rear_layer[:, 0:pad_size*2, 0:size_w+pad_size*4] = 0.0
gpu_rear_layer[:, 0:size_h+pad_size*4, 0:pad_size*2] = 0.0
gpu_rear_layer[:, 0:size_h+pad_size*4, size_w+pad_size*2:size_w+pad_size*4] = 0.0
gpu_rear_layer[:, size_h+pad_size*2:size_h+pad_size*4, 0:size_w+pad_size*4] = 0.0

bias = cupy.full((chanel_num, size_h, size_w), 0.00001, cupy.float32)




#Iterate until convergence
for iter in range(iteration):
	print('iter = %02d' %iter)

#Update front layer
	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_front_layer[:, pad_size+phi:pad_size+size_h+phi, pad_size+theta:pad_size+size_w+theta] * gpu_middle_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * gpu_rear_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta]

	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			numerator += gpu_light_field[:, phi*view_point_row_num+theta, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta] * gpu_middle_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta] * gpu_rear_layer[:, pad_size*4-phi*2:size_h+pad_size*4-phi*2, pad_size*4-theta*2:size_w+pad_size*4-theta*2]

			denominator += gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta] * gpu_middle_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta] * gpu_rear_layer[:, pad_size*4-phi*2:size_h+pad_size*4-phi*2, pad_size*4-theta*2:size_w+pad_size*4-theta*2]


	gpu_front_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_front_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * ((numerator[:, :, :]+bias[:, :, :])/(denominator[:, :, :]+bias[:, :, :]))

	gpu_front_layer[gpu_front_layer < 0.0] = 0.0
	gpu_front_layer[gpu_front_layer > 1.0] = 1.0

	numerator = 0.0
	denominator = 0.0


#Update middle layer
	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_front_layer[:, pad_size+phi:pad_size+size_h+phi, pad_size+theta:pad_size+size_w+theta] * gpu_middle_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * gpu_rear_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta]

	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			numerator += gpu_light_field[:, phi*view_point_row_num+theta, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * gpu_front_layer[:, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta] * gpu_rear_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta]

			denominator += gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * gpu_front_layer[:, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta] * gpu_rear_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta]


	gpu_middle_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_middle_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * ((numerator[:, :, :]+bias[:, :, :])/(denominator[:, :, :]+bias[:, :, :]))

	gpu_middle_layer[gpu_middle_layer < 0.0] = 0.0
	gpu_middle_layer[gpu_middle_layer > 1.0] = 1.0

	numerator = 0.0
	denominator = 0.0


#Update rear layer
	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_front_layer[:, pad_size+phi:pad_size+size_h+phi, pad_size+theta:pad_size+size_w+theta] * gpu_middle_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * gpu_rear_layer[:, pad_size*3-phi:size_h+pad_size*3-phi, pad_size*3-theta:size_w+pad_size*3-theta]

	for phi in range(view_point_row_num):
		for theta in range(view_point_row_num):
			numerator += gpu_light_field[:, phi*view_point_row_num+theta, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta] * gpu_front_layer[:, phi*2:size_h+phi*2, theta*2:size_w+theta*2] * gpu_middle_layer[:, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta]

			denominator += gpu_reproduced_light_field[:, phi*view_point_row_num+theta, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta] * gpu_front_layer[:, phi*2:size_h+phi*2, theta*2:size_w+theta*2] * gpu_middle_layer[:, pad_size+phi:size_h+pad_size+phi, pad_size+theta:size_w+pad_size+theta]


	gpu_rear_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] = gpu_rear_layer[:, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2] * ((numerator[:, :, :]+bias[:, :, :])/(denominator[:, :, :]+bias[:, :, :]))

	gpu_rear_layer[gpu_rear_layer < 0.0] = 0.0
	gpu_rear_layer[gpu_rear_layer > 1.0] = 1.0

	numerator = 0.0
	denominator = 0.0

		

gpu_front_layer = chainer.cuda.to_cpu(gpu_front_layer)
gpu_middle_layer = chainer.cuda.to_cpu(gpu_middle_layer)
gpu_rear_layer = chainer.cuda.to_cpu(gpu_rear_layer)

for ch in range(chanel_num):
	cpu_layer[0][:, :, ch] = gpu_front_layer[ch, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]
	cpu_layer[1][:, :, ch] = gpu_middle_layer[ch, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]
	cpu_layer[2][:, :, ch] = gpu_rear_layer[ch, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]

gpu_reproduced_light_field = chainer.cuda.to_cpu(gpu_reproduced_light_field)

for t in range(view_point_num):
	cpu_reproduced_light_field[t][:, :, 0] = gpu_reproduced_light_field[0, t, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]
	cpu_reproduced_light_field[t][:, :, 1] = gpu_reproduced_light_field[1, t, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]
	cpu_reproduced_light_field[t][:, :, 2] = gpu_reproduced_light_field[2, t, pad_size*2:size_h+pad_size*2, pad_size*2:size_w+pad_size*2]



#Save reproduced light field
mse = 0.0
for t in range(view_point_num):
	img_save = cpu_reproduced_light_field[t]
	img_save[img_save < 0.0] = 0.0
	img_save[img_save > 1.0] = 1.0
	img_err = cpu_light_field[t][pad_size:size_h-pad_size, pad_size:size_w-pad_size, :] - img_save[pad_size:size_h-pad_size, pad_size:size_w-pad_size, :]
	img_err = img_err**2
	mse += np.sum(img_err)
	img_save = Image.fromarray(np.uint8(img_save*255.0))

	img_save.save("Dst/reproducedimage/dst%02d.png" % (t+1))

mse /= img_err.shape[0]*img_err.shape[1]*img_err.shape[2]*view_point_num

print("psnr:%f"%(10*np.log10(1.0 / (mse))))


#Save layers
for t in range(layer_num):
	img_save = cpu_layer[t]
	img_save[img_save < 0.0] = 0.0
	img_save[img_save > 1.0] = 1.0
	img_save = Image.fromarray(np.uint8(img_save*255.0))
	img_save.save("Dst/layer/layer%d.png" % (t))

