from chainer import cuda, optimizers, Variable, serializers
import numpy as np
from PIL import Image
import os

from model import*


epoch_num = 20
batch_size = 15
chanel_num = 3
image_file_num = 7
block_num = 10
block_set_num = image_file_num*block_num
block_size = 64
light_kind_num = 6
train_data_num = chanel_num*block_set_num*light_kind_num
check_num = train_data_num
check_num = check_num - check_num%15

view_point_row_num = 5
view_point_num = view_point_row_num*view_point_row_num
pad_size = (int)(view_point_row_num/2)


aug_data = np.zeros((batch_size, view_point_num, block_size, block_size), np.float32)
cnt = 0
loss_sum = 0.0

xp = cuda.cupy
cuda.get_device(0).use()


if os.path.isdir("./Dst") == False:
	os.mkdir("./Dst")


if os.path.isdir("./Dst/Model") == False:
	os.mkdir("./Dst/Model")




#Load train datas
load_light_field = np.zeros((block_set_num*chanel_num, view_point_num, block_size, block_size), np.uint8)
start_index_R = 0;
start_index_G = image_file_num*block_num
start_index_B = image_file_num*block_num*2


for n in range(image_file_num):
	if(n % 10 == 0):
		print("load %03d" % (n))

	fname = "../../TrainingData/%03d.png" % (n)
	img_load = Image.open(fname)
	img_load = np.asarray(img_load)

	for b in range(block_num):
		for t in range(view_point_num):
			load_light_field[n*block_num+b +start_index_R,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,0]
			load_light_field[n*block_num+b +start_index_G,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,1]
			load_light_field[n*block_num+b +start_index_B,t,:,:] = img_load[block_size*b:block_size*b+block_size,block_size*t:block_size*t+block_size,2]




#Model initaition
calclayer = CalcLayer()
calclayer.to_gpu()
calclayer_optimizer = optimizers.Adam(alpha=0.0001, beta1=0.9, beta2=0.999, eps=1e-08)
calclayer_optimizer.setup(calclayer)



#Train loop
for i in range(1, epoch_num+1):

	sff = np.random.permutation(train_data_num)

	for n in range(0, train_data_num, batch_size):
		#Data augmentation by intensity levels
		for j in range(batch_size):
			aug_data[j,:,:,:] = ((1-(sff[n+j]%light_kind_num)/10)*load_light_field[int(sff[n+j]/light_kind_num)]/255).astype(np.float32)

		gray_light_field = cuda.to_gpu(aug_data)
		gray_light_field = Variable(gray_light_field)

		#Train models
		calclayer.cleargrads()

		gray_layer = calclayer(gray_light_field)
		gray_layer=F.expand_dims(gray_layer,2)
		gray_reproduced_light_field = gray_layer[:, 0, :, 0:block_size-pad_size*2, 0:block_size-pad_size*2]

		#Physical display simulation
		for p in range (view_point_row_num):
			for t in range (view_point_row_num):
				gray_reproduced_light_field = F.concat([gray_reproduced_light_field, gray_layer[:, 0, :, p:block_size+p-pad_size*2, t:block_size+t-pad_size*2] + gray_layer[:, 1, :, pad_size:block_size-pad_size, pad_size:block_size-pad_size] + gray_layer[:, 2, :, pad_size*2-p:block_size-p, pad_size*2-t:block_size-t]], axis=1)

		loss = F.mean_squared_error(gray_reproduced_light_field[:, 1:view_point_num+1, :, :], gray_light_field[:, :, pad_size:block_size-pad_size, pad_size:block_size-pad_size])

		loss.backward()
		calclayer_optimizer.update()

		print("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f" % (i, epoch_num, n, train_data_num, 10*xp.log10(1.0/loss.data)))

		loss_sum += loss.data
		cnt += 1


		#Save calculate average psnr
		if(n%check_num == 0):
			print("@@@    average psnr:%.6f    @@@" % (10*xp.log10(1.0 / (loss_sum / cnt))))
			f_psnr = open("./Dst/average_psnr.txt", "a")
			f_psnr.write("epoch:%04d/%04d, data:%06d/%06d, psnr:%.6f\n" % (i, epoch_num, n, train_data_num, 10*xp.log10(1.0 / (loss_sum / cnt))))
			f_psnr.close()

			cnt = 0
			loss_sum = 0.0
	


	#Save models every epoch
	calclayer.to_cpu()
	serializers.save_npz("./Dst/Model/calclayer%04d.npz" %i, calclayer)
	calclayer.to_gpu()

